model: fix spacing
[c11tester.git] / model.h
diff --git a/model.h b/model.h
index 62c762711bcb8300765483ee075ec9747d412d5f..081a8ce603bcba39728c2a30bf59ddb9c3261649 100644 (file)
--- a/model.h
+++ b/model.h
@@ -23,6 +23,7 @@ class CycleGraph;
 class Promise;
 class Scheduler;
 class Thread;
+struct model_snapshot_members;
 
 /** @brief Shorthand for a list of release sequence heads */
 typedef std::vector< const ModelAction *, ModelAlloc<const ModelAction *> > rel_heads_list_t;
@@ -36,23 +37,34 @@ struct model_params {
        int maxfuturedelay;
        unsigned int fairwindow;
        unsigned int enabledcount;
+       unsigned int bound;
+
+       /** @brief Maximum number of future values that can be sent to the same
+        *  read */
+       int maxfuturevalues;
+
+       /** @brief Only generate a new future value/expiration pair if the
+        *  expiration time exceeds the existing one by more than the slop
+        *  value */
+       unsigned int expireslop;
 };
 
 struct PendingFutureValue {
-       uint64_t value;
-       modelclock_t expiration;
-       ModelAction * act;
+       ModelAction *writer;
+       ModelAction *act;
 };
 
-/**
- * Structure for holding small ModelChecker members that should be snapshotted
- */
-struct model_snapshot_members {
-       ModelAction *current_action;
-       int next_thread_id;
-       modelclock_t used_sequence_numbers;
-       Thread *nextThread;
-       ModelAction *next_backtrack;
+/** @brief Records information regarding a single pending release sequence */
+struct release_seq {
+       /** @brief The acquire operation */
+       ModelAction *acquire;
+       /** @brief The head of the RMW chain from which 'acquire' reads; may be
+        *  equal to 'release' */
+       const ModelAction *rf;
+       /** @brief The head of the potential longest release sequence chain */
+       const ModelAction *release;
+       /** @brief The write(s) that may break the release sequence */
+       std::vector<const ModelAction *> writes;
 };
 
 /** @brief The central structure for model-checking */
@@ -72,11 +84,11 @@ public:
 
        void add_thread(Thread *t);
        void remove_thread(Thread *t);
-       Thread * get_thread(thread_id_t tid);
-       Thread * get_thread(ModelAction *act);
+       Thread * get_thread(thread_id_t tid) const;
+       Thread * get_thread(ModelAction *act) const;
 
        thread_id_t get_next_id();
-       int get_num_threads();
+       unsigned int get_num_threads() const;
        Thread * get_current_thread();
 
        int switch_to_master(ModelAction *act);
@@ -86,45 +98,48 @@ public:
        bool isfeasible();
        bool isfeasibleotherthanRMW();
        bool isfinalfeasible();
-       void check_promises(ClockVector *old_cv, ClockVector * merge_cv);
+       void check_promises_thread_disabled();
+       void mo_check_promises(thread_id_t tid, const ModelAction *write);
+       void check_promises(thread_id_t tid, ClockVector *old_cv, ClockVector * merge_cv);
        void get_release_seq_heads(ModelAction *act, rel_heads_list_t *release_heads);
        void finish_execution();
        bool isfeasibleprefix();
        void set_assert() {asserted=true;}
+       bool is_deadlocked() const;
 
        /** @brief Alert the model-checker that an incorrectly-ordered
         * synchronization was made */
        void set_bad_synchronization() { bad_synchronization = true; }
 
        const model_params params;
+       Scheduler * get_scheduler() { return scheduler;}
+       Node * get_curr_node();
 
        MEMALLOC
 private:
        /** The scheduler to use: tracks the running/ready Threads */
        Scheduler *scheduler;
 
+       bool sleep_can_read_from(ModelAction * curr, const ModelAction *write);
        bool thin_air_constraint_may_allow(const ModelAction * writer, const ModelAction *reader);
+       bool mo_may_allow(const ModelAction * writer, const ModelAction *reader);
        bool has_asserted() {return asserted;}
        void reset_asserted() {asserted=false;}
        int num_executions;
        int num_feasible_executions;
        bool promises_expired();
-
+       void execute_sleep_set();
+       void wake_up_sleeping_actions(ModelAction * curr);
        modelclock_t get_next_seq_num();
 
-       /**
-        * Stores the ModelAction for the current thread action.  Call this
-        * immediately before switching from user- to system-context to pass
-        * data between them.
-        * @param act The ModelAction created by the user-thread action
-        */
-       void set_current_action(ModelAction *act) { priv->current_action = act; }
+       void set_current_action(ModelAction *act);
        Thread * check_current_action(ModelAction *curr);
-       ModelAction * initialize_curr_action(ModelAction *curr);
+       bool initialize_curr_action(ModelAction **curr);
        bool process_read(ModelAction *curr, bool second_part_of_rmw);
        bool process_write(ModelAction *curr);
        bool process_mutex(ModelAction *curr);
        bool process_thread_action(ModelAction *curr);
+       void process_relseq_fixup(ModelAction *curr, work_queue_t *work_queue);
        bool check_action_enabled(ModelAction *curr);
 
        bool take_step();
@@ -137,6 +152,7 @@ private:
        void reset_to_initial_state();
        bool resolve_promises(ModelAction *curr);
        void compute_promises(ModelAction *curr);
+       void compute_relseq_breakwrites(ModelAction *curr);
 
        void check_curr_backtracking(ModelAction * curr);
        void add_action_to_lists(ModelAction *act);
@@ -148,9 +164,8 @@ private:
        void post_r_modification_order(ModelAction *curr, const ModelAction *rf);
        bool r_modification_order(ModelAction *curr, const ModelAction *rf);
        bool w_modification_order(ModelAction *curr);
-       bool release_seq_head(const ModelAction *rf, rel_heads_list_t *release_heads) const;
+       bool release_seq_heads(const ModelAction *rf, rel_heads_list_t *release_heads, struct release_seq *pending) const;
        bool resolve_release_sequences(void *location, work_queue_t *work_queue);
-       void do_complete_join(ModelAction *join);
 
        ModelAction *diverge;
        ModelAction *earliest_diverge;
@@ -167,25 +182,33 @@ private:
         * to a trace of all actions performed on the object. */
        HashTable<const void *, action_list_t, uintptr_t, 4> *lock_waiters_map;
 
+       /** Per-object list of actions. Maps an object (i.e., memory location)
+        * to a trace of all actions performed on the object. */
+       HashTable<const void *, action_list_t, uintptr_t, 4> *condvar_waiters_map;
+
        HashTable<void *, std::vector<action_list_t>, uintptr_t, 4 > *obj_thrd_map;
-       std::vector<Promise *> *promises;
-       std::vector<struct PendingFutureValue> *futurevalues;
+       std::vector< Promise *, SnapshotAlloc<Promise *> > *promises;
+       std::vector< struct PendingFutureValue, SnapshotAlloc<struct PendingFutureValue> > *futurevalues;
 
        /**
* List of acquire actions that might synchronize with one or more
* release sequence. Release sequences might be determined lazily as
* promises are fulfilled and modification orders are established. Each
* ModelAction in this list must be an acquire operation.
         */
-       std::vector<ModelAction *> *pending_acq_rel_seq;
+       std::vector< struct release_seq *, SnapshotAlloc<struct release_seq *> > *pending_rel_seqs;
 
-       std::vector<ModelAction *> *thrd_last_action;
+       std::vector< ModelAction *, SnapshotAlloc<ModelAction *> > *thrd_last_action;
        NodeStack *node_stack;
 
        /** Private data members that should be snapshotted. They are grouped
         * together for efficiency and maintainability. */
        struct model_snapshot_members *priv;
 
+       /** A special model-checker Thread; used for associating with
+        *  model-checker-related ModelAcitons */
+       Thread *model_thread;
+
        /**
         * @brief The modification order graph
         *
@@ -167,25 +182,33 @@ private:
 
        /**
-        * List of acquire actions that might synchronize with one or more
-        * release sequence. Release sequences might be determined lazily as
-        * promises are fulfilled and modification orders are established. Each
-        * ModelAction in this list must be an acquire operation.
+        * List of pending release sequences. Release sequences might be
+        * determined lazily as promises are fulfilled and modification orders
+        * are established. Each entry in the list may only be partially
+        * filled, depending on its pending status.
         */